import pygame

from enum import *

import render
import maps
import content
import input
import combat
import mobs
import font
import ui
import engine
import pathing

class Mode:
    DEFAULT = 0
    TARGET = 1
    ATTACK = 2
    SKILLPANE = 3
    POSSESSION = 4
    AI = 5
    STATSPANE = 6
    DEAD = 7
    
class Cursor:
    def __init__(self):
        self.visible = False
        self._source = (0, 0)
        self.range = 1
        self.x = 0
        self.y = 0
        self.func = None
        self.skill = None
        
    def tryMove(self, x, y):
        self.x += x
        self.y += y
        
        if abs(self.x - self._source[0]) > self.range:
            if self.x < self._source[0]:
                self.x = self._source[0] - self.range
            else:
                self.x = self._source[0] + self.range
        if abs(self.y - self._source[1]) > self.range:
            if self.y < self._source[1]:
                self.y = self._source[1] - self.range
            else:
                self.y = self._source[1] + self.range
                
    def setSource(self, value):
        self._source = value
        self.x = value[0]
        self.y = value[1]

    def getSource(self):
        return self._source
    
class Game(render.Renderable):
    def __init__(self):
        pygame.mixer.pre_init(44100, -16, 2, 2048)
        pygame.init()
        pygame.display.set_caption("MiniBoss")
        
        self.levels = []
        
        self.quitting = False
        self.input = input.InputHandler()
        self.renderer = render.Renderer(engine.scale)
        content.init()
        
        self.level = maps.generate(engine.roomSize, engine.floorWidth, engine.floorHeight)
        self.level.setDeathCallback(self.mobileDied)
        
        self.renderer.renderables.append(self.level.renderer)
        self.renderer.renderables.append(self)
        self.ui = ui.UI(self.renderer)
        
        self.bindKeys()
        
        self.playersTurn = True
        self.moveSpeed = 50
        
        self.mobiles = []
        self.moveElapsed = 0
        self.lastTick = 0
        
        self.mode = Mode.DEFAULT
        self.cursor = Cursor()
        self.target = None
        self.curSound = None
        self.soundElapsed = 0
        self.soundFunc = None
        
        self.activity = None
        
        engine.player.mobile.maxHealth += engine.player.health
        engine.player.mobile.str += engine.player.str
        engine.player.mobile.agi += engine.player.agi
        engine.player.mobile.con += engine.player.con
        engine.player.mobile.mag += engine.player.mag
        engine.player.mobile.wil += engine.player.wil
        engine.player.mobile.health = engine.player.mobile.maxHealth
        
        self.floor = 1
        self.levels.append(self.level)
        
        engine.log("Welcome to Mini Boss!")
        engine.log("Press 'L' to open or close this log.")
        engine.log("")
        engine.log("You are an imp, a weak creature with the ability to")
        engine.log("possess other creatures. You want to become the")
        engine.log("dungeon's final boss, who resides on floor 12.")
        engine.log("")
        engine.log("Possess creatures to learn their abilities, level up")
        engine.log("to add stat bonusses to any possessed creature.")
        engine.log("")
        engine.log("Numpad to move, ASD for attack, skill, and status.")
        engine.log("Space/numpad 5 to wait, F9/F10 to resize the window.")
        engine.log("")
        engine.log("")
        engine.log("")
        engine.log("")
        engine.log("")
        
    def bindKeys(self):
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP1, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP2, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP3, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP4, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP6, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP7, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP8, self.move)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP9, self.move)
        
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP5, self.wait)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_SPACE, self.waitOrConfirm)
        
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_a, self.beginAttack)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_KP_ENTER, self.confirm)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_RETURN, self.confirm)
        
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_s, self.openSkillPane)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_d, self.openStatsPane)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_l, self.showLog)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_ESCAPE, self.closeUI)

        self.input.attachHandler(pygame.KEYDOWN, pygame.K_F9, self.resizeWindow)
        self.input.attachHandler(pygame.KEYDOWN, pygame.K_F10, self.resizeWindow)
        
    def resizeWindow(self, event, key):
        lastScale = engine.scale
        if key == pygame.K_F9:
            engine.scale -= 1
        if key == pygame.K_F10:
            engine.scale += 1
            
        if engine.scale < 1:
            engine.scale = 1
        if engine.scale > 4:
            engine.scale = 4
            
        if engine.scale != lastScale:
            self.renderer.resize(engine.scale)
    
    def closeUI(self, event, key):
        if self.mode == Mode.SKILLPANE:
            self.ui.skillPane.visible = False
            self.mode = Mode.DEFAULT
        elif self.mode == Mode.STATSPANE:
            self.ui.statsPane.visible = False
            self.mode = Mode.DEFAULT
        elif self.mode == Mode.TARGET:
            self.cancelCursor()
        elif self.mode == Mode.POSSESSION:
            self.ui.possessPane.visible = False
            self.mode = Mode.DEFAULT
            self.ui.possessPane.mob.cleanUp()
            
    def showLog(self, event, key):
        if self.mode != Mode.DEFAULT or not self.playersTurn:
            return
        
        self.ui.textLog.switchStyle()
        
    def openStatsPane(self, event, key):
        if (self.mode != Mode.DEFAULT and not self.mode == Mode.STATSPANE) or not self.playersTurn:
            return
        
        if self.mode == Mode.STATSPANE:
            self.closeUI(None, None)
            return
        
        self.mode = Mode.STATSPANE
        self.ui.statsPane.visible = True
        
    def openSkillPane(self, event, key):
        if (self.mode != Mode.DEFAULT and not self.mode == Mode.SKILLPANE) or not self.playersTurn:
            return
        
        if self.mode == Mode.SKILLPANE:
            self.closeUI(None, None)
            return
        
        self.mode = Mode.SKILLPANE
        self.ui.skillPane.visible = True
   
    def beginAttack(self, event, key):
        if self.mode != Mode.DEFAULT or not self.playersTurn:
            return
        
        self.beginTarget(self.attack)
        
    def beginTarget(self, func, source=None, skill=None):
        self.mode = Mode.TARGET
        self.cursor.visible = True
        
        self.cursor.func = func
        if source == None:
            source = engine.player.mobile.getPosition()
        self.cursor.setSource(source)
        self.cursor.skill = skill
        if self.cursor.skill != None:
            self.cursor.range = skill.range
        else:
            self.cursor.range = 1
       
    def attack(self):
        self.cursor.visible = False
        mob = self.level.mobileAt(self.cursor.x, self.cursor.y)
        
        if mob == None or \
        (mob == engine.player.mobile and self.cursor.skill == None) or \
        (mob == engine.player.mobile and not self.cursor.skill.canTargetSelf):
            self.mode = Mode.DEFAULT
            return
        
        self.mode = Mode.ATTACK
        self.target = mob
        self.source = engine.player.mobile
        
        sound = content.getSound("swing")
        if self.cursor.skill != None:
            if self.cursor.skill.swingSound != None:
                sound = content.getSound(self.cursor.skill.swingSound)
            else:
                sound = None
        
        self.playSoundAndWait(sound, self.resolvePlayerAttack)
        
    def resolveAttack(self, source, target, skill=None):
        if skill == None:
            dmg = combat.resolvePhysicalAttack(source, target)
            if dmg != None:
                self.playSoundAndWait(content.getSound("bump"), self.endAttack)
            else:
                self.endAttack()
        else:
            sound = combat.resolveSkill(skill, source, target)
            if sound != None:
                self.playSoundAndWait(content.getSound(sound), self.endAttack)
            else:
                self.endAttack()
                
    def resolvePlayerAttack(self):
        self.resolveAttack(self.source, self.target, self.cursor.skill)
                
    def mobileAttack(self, activity):
        self.activity = activity
        
        sound = "swing"
        if self.activity.skill != None:
            sound = self.activity.skill.swingSound
        
        self.playSoundAndWait(content.getSound(sound), self.resolveMobileAttack)
        
    def resolveMobileAttack(self):
        self.resolveAttack(self.activity.source, self.activity.target, self.activity.skill)
            
    def endAttack(self):
        if self.mode == Mode.ATTACK:
            self.mode = Mode.DEFAULT
            self.endTurn()
        elif self.mode == Mode.AI:
            self.activity = None
        
    def playSoundAndWait(self, sound, func):
        if sound == None:
            func()
            return
        self.soundElapsed = 0
        self.curSound = sound
        engine.playSound(self.curSound)
        self.soundFunc = func
        
    def confirmTarget(self, event, key):
        if self.cursor.visible:
            self.cursor.func()
            
    def confirmSkillUse(self):
        skill = self.ui.skillPane.getSelectedSkill()
        if engine.player.sp < skill.cost or skill.passive:
            engine.playSound(content.getSound("nope"))
            return
        self.closeUI(None, None)
        self.beginTarget(self.useSkill, skill=skill)
        
    def useSkill(self):
        self.cursor.visible = False
        engine.player.sp -= self.cursor.skill.cost
        self.attack()
        
    def confirmStatPoint(self):
        if engine.player.statPoints <= 0:
            return
        
        pos = self.ui.statsPane.cursorPos
        
        if pos == 0:
            engine.player.health += 1
            engine.player.mobile.health += 1
            engine.player.mobile.maxHealth += 1
        elif pos == 1:
            engine.player.str += 1
            engine.player.mobile.str += 1
        elif pos == 2:
            engine.player.agi += 1
            engine.player.mobile.agi += 1
        elif pos == 3:
            engine.player.con += 1
            engine.player.mobile.con += 1
        elif pos == 4:
            engine.player.mag += 1
            engine.player.mobile.mag += 1
        elif pos == 5:
            engine.player.wil += 1
            engine.player.mobile.wil += 1
            
        engine.player.statPoints -= 1
        engine.playSound(content.getSound("uiping"))
        
    def confirm(self, event, key):
        if self.mode == Mode.TARGET:
            self.confirmTarget(event, key)
        elif self.mode == Mode.SKILLPANE:
            self.confirmSkillUse()
        elif self.mode == Mode.STATSPANE:
            self.confirmStatPoint()
        elif self.mode == Mode.POSSESSION:
            self.possess(self.ui.possessPane.mob)
        
    def waitOrConfirm(self, event, key):
        if self.mode == Mode.DEFAULT:
            self.wait(event, key)
        else:
            self.confirm(event, key)
    
    def wait(self, event, key):
        if self.mode != Mode.DEFAULT:
            return
        self.endTurn()
        
    def cancelCursor(self):
        self.cursor.visible = False
        self.mode = Mode.DEFAULT
    
    def moveCursor(self, event, key):
        if key == pygame.K_KP2:
            self.cursor.tryMove(0, 1)
        elif key == pygame.K_KP8:
            self.cursor.tryMove(0, -1)
        elif key == pygame.K_KP6:
            self.cursor.tryMove(1, 0)
        elif key == pygame.K_KP4:
            self.cursor.tryMove(-1, 0)
        elif key == pygame.K_KP7:
            self.cursor.tryMove(-1, -1)
        elif key == pygame.K_KP9:
            self.cursor.tryMove(1, -1)
        elif key == pygame.K_KP1:
            self.cursor.tryMove(-1, 1)
        elif key == pygame.K_KP3:
            self.cursor.tryMove(1, 1)
                    
    def move(self, event, key):
        if not self.playersTurn:
            return
        
        if self.mode == Mode.TARGET:
            self.moveCursor(event, key)
            return
        elif self.mode == Mode.SKILLPANE:
            self.ui.skillPane.moveCursor(event, key)
            return
        elif self.mode == Mode.STATSPANE:
            if engine.player.statPoints > 0:
                self.ui.statsPane.moveCursor(event, key)
            return
        elif self.mode == Mode.POSSESSION:
            self.ui.possessPane.moveCursor(event, key)
            return
        elif self.mode != Mode.DEFAULT:
            return
        
        res = True
        if key == pygame.K_KP2:
            res = self.level.movePlayer(Direction.DOWN)
        elif key == pygame.K_KP8:
            res = self.level.movePlayer(Direction.UP)
        elif key == pygame.K_KP6:
            res = self.level.movePlayer(Direction.RIGHT)
        elif key == pygame.K_KP4:
            res = self.level.movePlayer(Direction.LEFT)
        elif key == pygame.K_KP7:
            res = self.level.movePlayer(Direction.UPLEFT)
        elif key == pygame.K_KP9:
            res = self.level.movePlayer(Direction.UPRIGHT)
        elif key == pygame.K_KP1:
            res = self.level.movePlayer(Direction.DOWNLEFT)
        elif key == pygame.K_KP3:
            res = self.level.movePlayer(Direction.DOWNRIGHT)
            
        if not res:
            engine.playSound(content.getSound("bump"))
        else:
            tile = self.level.getTile(engine.player.mobile.x, engine.player.mobile.y)
            if tile.type == TileType.STAIRSDOWN or tile.type == TileType.STAIRSUP:
                self.climbStairs(tile.type)
            else:
                self.endTurn()
                
    def climbStairs(self, type):
        if type == TileType.STAIRSDOWN:
            self.floor += 1
        else:
            self.floor -= 1
            
        if len(self.levels) < self.floor:
            newLevel = maps.generate(engine.roomSize, engine.floorWidth, engine.floorHeight, self.floor * 4, False, self.floor)
            self.levels.append(newLevel)
            
        self.renderer.renderables.remove(self.level.renderer)
        engine.player.mobile.cleanUp()
        
        self.level = self.levels[self.floor - 1]
        self.level.addMobile(engine.player.mobile)
        self.level.track(engine.player.mobile)
        self.mobiles = list(self.level.mobiles)
        
        self.level.setDeathCallback(self.mobileDied)
        self.renderer.renderables.insert(0, self.level.renderer)
        
        if type == TileType.STAIRSDOWN:
            engine.player.mobile.moveTo(self.level.stairsUp[0], self.level.stairsUp[1])
        else:
            engine.player.mobile.moveTo(self.level.stairsDown[0], self.level.stairsDown[1])
            
        self.ui.bottomPane.floor = self.floor;
        
    def endTurn(self):
        self.playersTurn = False
        self.mobiles = list(self.level.mobiles)
        self.moveElapsed = 0
        self.mode = Mode.AI
        
    def beginTurn(self):
        if self.mode == Mode.DEAD:
            return
        
        self.playersTurn = True
        engine.player.regenSp()
        self.moveElapsed = 0
        self.mode = Mode.DEFAULT
        
        for mobile in self.level.mobiles:
            mobile.updateStatus()
        
        if engine.player.mobile.hasCondition(Condition.POISON):
            engine.playSound(content.getSound("envenom"))
            engine.log("Poisoned!")
        
    def mobileDied(self, mob, source):
        skill = mob.hasSkill("Reanimate")
        if skill != None and mob.health <= 0:
            name = engine.getName(mob)
            sound = skill.resolve(mob, mob)
            if sound != None:
                engine.playSound(content.getSound(sound))
            if not mob.dead:
                return
        
        if mob == engine.player.mobile:
            engine.log("Your form has dissipated")
            
            if engine.player.mobile.getName() == "Imp":
                self.mode = Mode.DEAD
                self.ui.deathScreen.visible = True
                return

            imp = mobs.Imp(engine.player.mobile.x, engine.player.mobile.y)
            
            imp.maxHealth += engine.player.health
            imp.str += engine.player.str
            imp.agi += engine.player.agi
            imp.con += engine.player.con
            imp.mag += engine.player.mag
            imp.wil += engine.player.wil
            
            if mob.health > 0:
                perc = 1.0 * mob.health / mob.maxHealth
            else:
                perc = 0.05
                
            imp.health = max(int(round(perc * imp.maxHealth)), 1)
            
            engine.player.mobile = imp
            self.level.addMobile(imp)
            self.level.track(imp)
        elif source == engine.player.mobile:
            engine.log("%s has died!" % mob.getName())
            
            engine.player.gainExpFromMobile(mob)
            
            if mob.hasCondition(Condition.VESSEL):
                self.beginPossess(mob)

    def beginPossess(self, target):
        self.ui.possessPane.choice = True
        self.mode = Mode.POSSESSION
        self.ui.possessPane.visible = True
        self.ui.possessPane.mob = target
        target.dead = False
        
    def possess(self, target):
        if self.ui.possessPane.choice:
            engine.log("You have possessed %s" % target.getName())
            engine.player.mobile.cleanUp()
            
            target.maxHealth += engine.player.health
            target.str += engine.player.str
            target.agi += engine.player.agi
            target.con += engine.player.con
            target.mag += engine.player.mag
            target.wil += engine.player.wil
            
            target.health = target.maxHealth
            target.ai = None
            engine.player.mobile = target
            self.level.track(engine.player.mobile)
        else:
            target.cleanUp()
            
        self.ui.possessPane.visible = False
        self.mode = Mode.DEFAULT
        self.endTurn()
    
    def run(self):
        while True:
            events = pygame.event.get()
            
            for event in events:
                if event.type == pygame.QUIT:
                    self.quitting = True
                    break
                
                if event.type == pygame.KEYDOWN or event.type == pygame.KEYUP:
                    self.input.tryHandle(event.type, event.key)
                    
            if self.quitting:
                break
            
            curTick = pygame.time.get_ticks()
            elapsed = curTick - self.lastTick;
            self.lastTick = curTick
            
            self.update(elapsed)
            self.beginDraw(elapsed)
            
    def waitOnSound(self, sound, elapsed):
        duration = sound.get_length() * 1000
        self.soundElapsed += elapsed
        if self.soundElapsed > duration:
            self.curSound = None
            self.soundFunc()
            
    def update(self, elapsed):
        if self.curSound != None:
            self.waitOnSound(self.curSound, elapsed)
            return
        
        if not self.playersTurn:
            self.moveElapsed += elapsed
            
            activity = None
            while self.moveElapsed >= self.moveSpeed:
                self.moveElapsed -= self.moveSpeed
                
                while activity == None:
                    mob = self.popMobile()
                    if mob == None:
                        break
                    
                    if not mob.dead:
                        if self.level.distanceToPlayer(mob) > engine.roomSize:
                            mob.update()
                        else:
                            activity = mob.update()
                        
                if activity != None and activity.type == ActivityType.ATTACK:
                    self.mobileAttack(activity)
                    return
                
            if len(self.mobiles) == 0:
                self.beginTurn()
                
    def popMobile(self):
        if len(self.mobiles) > 0:
            return self.mobiles.pop(0)
        else:
            return None
    
    def beginDraw(self, elapsed):
        self.renderer.render()
        
    def render(self, screen):
        if self.cursor.visible:
            drawPos = self.level.renderer.camera.translate(self.cursor.x, self.cursor.y)
            screen.blit(content.getImage("reticule"), drawPos)

game = Game()
engine.setGameRef(game)
game.run()