import weakref
import operator

from enum import *

import engine
import maps
import content
import ai
import skills
    
hazardRatings = { "PlagueRat": 1,
                 "VampireBat": 2,
                 "Skeleton": 3,
                 "Mortipede": 4,
                 "Ghost": 5,
                 "Sorceror": 6,
                 "Minotaur": 7,
                 "HellRat": 8,
                 "EtherBat": 10,
                 "Bones": 12,
                 "RotWorm": 14,
                 "HungryGhost": 16,
                 "Liche": 18,
                 "Megataur": 20,
                 "Balrog": 35 }
sortedHazardRatings = sorted(hazardRatings.iteritems(), key=operator.itemgetter(1))

class Mobile:
    def __init__(self, x, y, maxHealth):
        self.maxHealth = maxHealth
        self.health = self.maxHealth
        self.str = 1
        self.agi = 1
        self.con = 1
        self.mag = 1
        self.wil = 1
        
        self.x = x
        self.y = y
        
        self.aggressive = False
        self.hazard = 0
        
        self.world = None
        self.img = None
        self.ai = None
        
        self.conditions = []
        self.onDeath = None
        self.dead = False
        
        self.skills = []
        
        self.poisonStrength = 0
        self.poisonSource = None
        self.poisonedTime = 0
        self.poisonDuration = 0
        
        self.fearStrength = 0
        self.fearTime = 0
        self.fearDuration = 0
        
        self.disposed = False
        
    def fearDebuff(self, strength):
        self.str -= strength
        self.agi -= strength
        self.con -= strength
        self.mag -= strength
        self.wil -= strength
        
    def setFearStrength(self, strength):
        if self.hasCondition(Condition.FEAR):
            self.fearDebuff(-self.fearStrength)
        
        self.fearStrength = strength
        self.fearDebuff(self.fearStrength)
                
    def heal(self, amount, showMsg=True):
        self.health += amount
        if self.health > self.maxHealth:
            self.health = self.maxHealth
        
        if showMsg:
            name = engine.getName(self)
            if name == "You":
                engine.log("You were healed for %s points" % amount)
            else:
                engine.log("%s was healed for %s points" % (name, amount))
        
    def addSkill(self, skill):
        if not skill in self.skills:
            self.skills.append(skill)
        
    def hasCondition(self, condition):
        return condition in self.conditions
    
    def moveTo(self, x, y):
        if self.world().checkCanMove(x, y) and self.checkCanMove(self.world().getTile(x, y)):
            self.x = x
            self.y = y
            return True
        else:
            return False
        
    def move(self, direction):
        newPos = [self.x, self.y]
        
        if direction == Direction.DOWN:
            newPos[1] += 1
        elif direction == Direction.UP:
            newPos[1] -= 1
        elif direction == Direction.RIGHT:
            newPos[0] += 1
        elif direction == Direction.LEFT:
            newPos[0] -= 1
        elif direction == Direction.UPLEFT:
            newPos[0] -= 1
            newPos[1] -= 1
        elif direction == Direction.UPRIGHT:
            newPos[0] += 1
            newPos[1] -= 1
        elif direction == Direction.DOWNLEFT:
            newPos[0] -= 1
            newPos[1] += 1
        elif direction == Direction.DOWNRIGHT:
            newPos[0] += 1
            newPos[1] += 1
        
        return self.moveTo(newPos[0], newPos[1])
        
    def checkCanMove(self, tile):
        if tile.type == TileType.FLOOR or tile.type == TileType.STAIRSDOWN or tile.type == TileType.STAIRSUP:
            return True
        else:
            return False
            
    def setWorld(self, world):
        self.world = weakref.ref(world)
        
    def removeWorld(self):
        self.world = None
        
    def removeCondition(self, condition):
        self.conditions.remove(condition)
        
    def update(self):
        if self.ai != None:
            return self.ai.tick()
        
    def updateStatus(self):
        regen = self.hasSkill("Regeneration")
        if regen != None:
            regen.resolve(self, self)
        
        if self.hasCondition(Condition.POISON):
            self.poisonedTime += 1
            if self.poisonedTime >= self.poisonDuration:
                self.removeCondition(Condition.POISON)
            else:
                self.inflictDamage(self.poisonStrength, self.poisonSource)
        
        if self.hasCondition(Condition.FEAR):
            self.fearTime += 1
            if self.fearTime >= self.fearDuration:
                self.removeCondition(Condition.FEAR)
                self.fearDebuff(-self.fearStrength)
                
    def getPosition(self):
        return (self.x, self.y)
    
    def inflictDamage(self, dmg, source):
        if self.world == None:
            print "inflicted damage on a being with no world"
            return   

        if source == engine.player.mobile:
            self.ai = ai.Aggressive(self, engine.roomSize * 2)
            self.ai.aggro = True
            
        skill = self.hasSkill("Ethereal")
        if skill != None:
            if skill.resolve(self, self) != None:
                dmg = 0
        
        self.health -= dmg
        if self.health <= 0:
            self.setDead(source)
            
    def cleanUp(self):
        self.world().removeMobile(self)
        self.removeWorld()
        self.disposed = True
        
    def setDead(self, source):
        self.dead = True
        if self.world == None and self.disposed:
            print "Mobile was set to dead, but world was None"
        else:
            self.world().mobileDied(self, source)
        if not self.dead:
            return
        self.cleanUp()
        
    def hasSkill(self, name):
        for skill in self.skills:
            if skill.name == name:
                return skill
        return None
                        
    def getName(self):
        return self.__class__.__name__

class Imp(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 20)
        self.str = 3
        self.agi = 4
        self.con = 1
        self.mag = 3
        self.wil = 5
        
        self.img = content.getImage("imp")
        
class PlagueRat(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 5)
        self.str = 3
        self.agi = 4
        self.con = 1
        
        self.hazard = hazardRatings["PlagueRat"]
        self.img = content.getImage("rat")
        self.ai = ai.Passive(self)
        
        self.addSkill(skills.Envenom())
           
class HellRat(Mobile): #5
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 15)
        self.str = 8
        self.agi = 13
        self.con = 6
        self.mag = 10
        self.wil = 5
        
        self.hazard = hazardRatings["HellRat"]
        self.img = content.getImage("hellrat")
        self.ai = ai.Passive(self)
        
        self.addSkill(skills.Envenom())
        self.addSkill(skills.Fireball())
               
class VampireBat(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 6)
        self.str = 3
        self.agi = 6
        self.con = 1
        self.mag = 4
        
        self.hazard = hazardRatings["VampireBat"]
        self.img = content.getImage("bat")
        self.ai = ai.Passive(self)
        
        self.addSkill(skills.LifeLeech())
          
class EtherBat(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 15)
        self.str = 6
        self.agi = 15
        self.con = 5
        self.mag = 12
        self.wil = 12
        
        self.hazard = hazardRatings["EtherBat"]
        self.img = content.getImage("etherbat")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.LifeLeech())
        self.addSkill(skills.MagicMissile())
        self.addSkill(skills.Ethereal())
         
class Skeleton(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 10)
        self.str = 5
        self.agi = 7
        self.con = 3
        
        self.hazard = hazardRatings["Skeleton"]
        self.img = content.getImage("skeleton")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.Reanimate())
               
class Bones(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 17)
        self.str = 17
        self.agi = 17
        self.con = 17
        self.mag = 1
        self.wil = 4
        
        self.hazard = hazardRatings["Bones"]
        self.img = content.getImage("bones")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.Reanimate())
        self.addSkill(skills.Charge())
        self.addSkill(skills.Regeneration())
        self.addSkill(skills.Chop())

class Mortipede(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 15)
        self.str = 5
        self.agi = 7
        self.con = 3
        self.wil = 7
        
        self.hazard = hazardRatings["Mortipede"]
        self.img = content.getImage("mortipede")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.Envenom())
        self.addSkill(skills.Fear())
               
class RotWorm(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 20)
        self.str = 25
        self.agi = 25
        self.con = 9
        self.mag = 1
        self.wil = 6
        
        self.hazard = hazardRatings["RotWorm"]
        self.img = content.getImage("rotworm")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.Envenom())
        self.addSkill(skills.Fear())
        self.addSkill(skills.Reanimate())
        self.addSkill(skills.Charge())

class Ghost(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 15)
        self.str = 6
        self.agi = 8
        self.con = 6
        
        self.hazard = hazardRatings["Ghost"]
        self.img = content.getImage("ghost")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.Reanimate())
        self.addSkill(skills.Ethereal())
               
class HungryGhost(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 20)
        self.str = 35 
        self.agi = 20
        self.con = 10
        self.mag = 1
        self.wil = 10
        
        self.hazard = hazardRatings["HungryGhost"]
        self.img = content.getImage("hungryghost")
        self.ai = ai.Aggressive(self)
        
        self.addSkill(skills.Reanimate())
        self.addSkill(skills.Ethereal())
        self.addSkill(skills.Regeneration())
        self.addSkill(skills.Fear())
               
class Sorceror(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 12)
        self.str = 2
        self.agi = 5
        self.con = 3
        self.mag = 12
        self.wil = 8
        
        self.hazard = hazardRatings["Sorceror"]
        self.img = content.getImage("sorceror")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.LifeLeech())
        self.addSkill(skills.Fireball())
        self.addSkill(skills.MagicMissile())

class Liche(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 20)
        self.str = 1
        self.agi = 1
        self.con = 8
        self.mag = 50
        self.wil = 20
        
        self.hazard = hazardRatings["Liche"]
        self.img = content.getImage("liche")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.LifeLeech())
        self.addSkill(skills.Fireball())
        self.addSkill(skills.MagicMissile())
        self.addSkill(skills.Ethereal())
        self.addSkill(skills.Reanimate())
        self.addSkill(skills.Regeneration())

class Minotaur(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 17)
        self.str = 10
        self.agi = 7
        self.con = 11
        self.mag = 1
        self.wil = 1
        
        self.hazard = hazardRatings["Minotaur"]
        self.img = content.getImage("minotaur")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.Charge())
        self.addSkill(skills.Regeneration())
   
class Megataur(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 35)
        self.str = 35
        self.agi = 25
        self.con = 35
        self.mag = 5
        self.wil = 20
        
        self.hazard = hazardRatings["Megataur"]
        self.img = content.getImage("megataur")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.Charge())
        self.addSkill(skills.Chop())
        self.addSkill(skills.Regeneration())
        
class Balrog(Mobile):
    def __init__(self, x, y):
        Mobile.__init__(self, x, y, 80)
        self.str = 53
        self.agi = 41
        self.con = 47
        self.mag = 43
        self.wil = 51
        
        self.hazard = hazardRatings["Balrog"]
        self.img = content.getImage("balrog")
        self.ai = ai.Mage(self)
        
        self.addSkill(skills.Charge())
        self.addSkill(skills.Chop())
        self.addSkill(skills.Regeneration())
        self.addSkill(skills.Envenom())
        self.addSkill(skills.Fear())
        self.addSkill(skills.LifeLeech())
        self.addSkill(skills.Fireball())
        self.addSkill(skills.MagicMissile())
        self.addSkill(skills.Ethereal())
        self.addSkill(skills.Reanimate())

def spawnByName(name, x, y):
    if name == "PlagueRat":
        return PlagueRat(x, y)
    elif name == "Skeleton":
        return Skeleton(x, y)
    elif name == "VampireBat":
        return VampireBat(x, y)
    elif name == "Mortipede":
        return Mortipede(x, y)
    elif name == "Ghost":
        return Ghost(x, y)
    elif name == "Sorceror":
        return Sorceror(x, y)
    elif name == "Minotaur":
        return Minotaur(x, y)
    elif name == "HellRat":
        return HellRat(x, y)
    elif name == "Bones":
        return Bones(x, y)
    elif name == "EtherBat":
        return EtherBat(x, y)
    elif name == "RotWorm":
        return RotWorm(x, y)
    elif name == "HungryGhost":
        return HungryGhost(x, y)
    elif name == "Liche":
        return Liche(x, y)
    elif name == "Megataur":
        return Megataur(x, y)
    elif name == "Balrog":
        return Balrog(x, y)